﻿/* NiuTrans.Tensor - an open-source tensor library
 * Copyright (C) 2018, Natural Language Processing Lab, Northestern University.
 * All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * $Created by: XIAO Tong (xiaotong@mail.neu.edu.cn) 2018-07-10
 */

#include <stdio.h>
#include "XNet.h"
#include "../tensor/XUtility.h"
#include "../tensor/function/FHeader.h"
#include "../tensor/core/CHeader.h"
#include "../tensor/test/Test.h"
#include "../sample/fnnlm/FNNLM.h"
#include "../sample/transformer/Transformer.h"
#include <time.h>

//#define CRTDBG_MAP_ALLOC
//#include <stdlib.h>
//#include <crtdbg.h>

void MatrixMul2DFloat16Test();
void MatrixMulFloat16AndFloatTest();
void ConvertTest();
void ConvertBackwardTest();
void DropoutFP16Test();
void SumFP16Test();
void SubFP16Test();
void MultiplyFP16Test();
void MultiplyDimFP16Test();
void UnsqueezeFP16Test();
void UnsqueezeFP16Test();
void SoftmaxBackwardFP16Test();
void ReduceMaxFP16Test();
void DropoutFP16BackwardTest();
void ScaleAndShiftFP16BackwardTest();
void SmallAttentionTest();
void LogSoftmaxFP16Test();
void TimeTest1();
void FloatToFloat16Test();
void FloatToInt8Test();
void SetDataFP16Test();
void SignFP16Test();
void MycublasGemmExTest();
void InitCPUFP16Test();
void SumDimFP16Test();
void SubDimFP16Test();
void DivDimFP16Test();
void DivFP16Test();
void ScaleAndShiftFP16Test();
void PowerFP16Test();
void ClipFP16Test();
void GatherFP16Test();
void SetDataGPUFP16Test();
void SumIntTest();

using namespace nts;
using namespace fnnlm;
using namespace transformer;

int main( int argc, const char ** argv )
{

    //MemTest();
    //return 0;
    //SplitBackwardTest();
    //return 0;
    //_CrtSetBreakAlloc(896);
    //BackwardTest();
    //return 0;
    //MatrixMul2DFloat16Test();
    //return 0;
    //MatrixMulFloat16AndFloatTest();
    //return 0;
    //ConvertTest();
    //return 0;  
    //ConvertBackwardTest();
    //return 0;
    //UnsqueezeFP16Test();
    //return 0;
    //ReduceMaxFP16Test();
    //return 0;
    //SoftmaxBackwardFP16Test();
    //return 0;
    //MultiplyDimFP16Test();
    //return 0;
    //DropoutFP16BackwardTest();
    //return 0;
    //ScaleAndShiftFP16BackwardTest();
    //return 0;
    //SmallAttentionTest();
    //return 0;
    //TimeTest1();
    //return 0;
    //SubFP16Test();
    //return 0;
    //LogSoftmaxFP16Test();
    //return 0;
    //FloatToFloat16Test();
    //return 0;
    //SetDataFP16Test();
    //return 0;
    //FloatToInt8Test();
    //return 0;
    //SignFP16Test();
    //return 0; 
    //SumDimFP16Test();
    //return 0;
    //SubDimFP16Test();
    //return 0;
    //DivDimFP16Test();
    //return 0;
    //MultiplyFP16Test();
    //return 0;
    //PowerFP16Test();
    //return 0;
    //DivFP16Test();
    //return 0;
    //SumFP16Test();
    //return 0;
    //ScaleAndShiftFP16Test();
    //return 0;
    //ClipFP16Test();
    //return 0;
    //GatherFP16Test();
    //return 0;

    //InitCPUFP16Test();
    //return 0;

    //MycublasGemmExTest();
    //return 0;

    //SumIntTest();
    //return 0;

    //DropoutFP16Test();
    //return 0;

    //SetDataGPUFP16Test();
    //return 0;


    if(argc > 1 && !strcmp(argv[1], "-fnnlm"))
        FNNLMMain(argc - 1, argv + 1);
    else if(argc > 1 && !strcmp(argv[1], "-t2t"))
        TransformerMain(argc - 1, argv + 1);
    else{
        fprintf(stderr, "Thanks for using NiuTrans.Network! This is a library for building\n");
        fprintf(stderr, "neural networks in an easy way. \n\n");
        fprintf(stderr, "Run this program with \"-test\" for unit test!\n");
        fprintf(stderr, "Or run this program with \"-fnnlm\" for sample FNNLM!\n");
    }

    //_CrtDumpMemoryLeaks();
    
    return 0;
}

void InitCPUFP16Test() {
    XTensor a;
    InitTensor2D(&a, 2, 2, X_FLOAT16, 0);
    a.SetDataRand(-3.0F, 3.0F);
    a.Dump(&a, stderr, "a:");
}

void MycublasGemmExTest() {

    XTensor a;
    XTensor a1;
    XTensor int8a;
    XTensor b;
    XTensor b1;
    XTensor int8b;
    XTensor c;
    XTensor c1;
    XTensor intc;
    InitTensor2D(&a, 1, 1, X_FLOAT, 0);
    InitTensor2D(&b, 1, 1, X_FLOAT, 0);
    InitTensor2D(&c, 1, 1, X_FLOAT, 0);
    InitTensor2D(&c1, 1, 1, X_INT, 0);
    a.SetDataRand(2.0F, 2.0F);
    b.SetDataRand(2.0F, 2.0F);
    int8a = ConvertDataType(a, X_INT8);
    int8b = ConvertDataType(b, X_INT8);

    a1 = ConvertDataType(int8a, X_FLOAT);
    b1 = ConvertDataType(int8b, X_FLOAT);
    a1.Dump(stderr, "a1:");
    b1.Dump(stderr, "b1:");

    _MatrixMul2D(&int8a, X_NOTRANS, &int8b, X_NOTRANS, &c, 1.0F, 0.0F);
    c.Dump(stderr, "c:");

    _MatrixMul2D(&int8a, X_NOTRANS, &int8b, X_NOTRANS, &c1, 1.0F, 0.0F);
    c1.Dump(stderr, "c1:");
}


void SumIntTest() {
    XTensor a;
    XTensor b;
    XTensor c;
    XTensor inta;
    XTensor intb;
    XTensor intc;
    InitTensor2D(&a, 2, 2, X_FLOAT, 0);
    InitTensor2D(&b, 2, 2, X_FLOAT, 0);
    a.SetDataRand(-5.0, 5.0);
    b.SetDataRand(-5.0, 5.0);

    a.Dump(stderr, "a:");
    b.Dump(stderr, "b:");  

    inta = ConvertDataType(a, X_INT);
    intb = ConvertDataType(b, X_INT);

    inta.Dump(stderr, "inta:");
    intb.Dump(stderr, "intb:");

    intc = Sum(inta, intb);

    intc.Dump(stderr, "intc:");

}

void SetDataGPUFP16Test() {

    srand(time(NULL));

    /*XTensor m;
    InitTensor2D(&m, 2, 2, X_FLOAT, 0);
    m.SetDataRand(0.0, 10.0);*/

    XTensor * m = NewTensor2D(2, 2, X_FLOAT, 0);
    m->SetDataRand(0.0, 10.0);

    //XTensor a1;
    //InitTensor2D(&a1, 2, 2, X_FLOAT, 0);
    //_CopyValues(&m, &a1);
    //_SetDataRand(&a1, -1.0F, 1.0F);
    //a1.Dump(&a1, stderr, "a:\n");

    /*XTensor a;
    InitTensor2D(&a, 2, 2, X_FLOAT16, 0);*/

    XTensor * a = NewTensor2D(2, 2, X_FLOAT16, 0);

    _ConvertDataType(m, a);

    a->Dump(a, stderr, "a:\n");

    _SetDataRand(a, 0.0F, 1.0F);

    a->Dump(a, stderr, "a:\n");

    //XTensor b;
    //InitTensor2D(&b, 2, 2, X_FLOAT, 0);
    //b = ConvertDataType(a, X_FLOAT);
    //b.Dump(stderr, "b:\n");

}

void ClipFP16Test() {
    XTensor a;
    InitTensor2D(&a, 2, 2, X_FLOAT16, 0);
    _SetDataRand(&a, -5.0F, 5.0F);

    a.Dump(&a, stderr, "a:\n");

    XTensor b;
    InitTensor2D(&b, 2, 2, X_FLOAT16, 0);
    b = Clip(a, -2.0F, 2.0F);
    b.Dump(&b, stderr, "b:\n");
}

void GatherFP16Test() {
    XTensor a;
    XTensor b;
    XTensor srcIndex;
    XTensor halfA;
    XTensor halfB;
    XTensor c;

    InitTensor1D(&srcIndex, 2, X_INT, 0);
    int m = 0;
    int n = 1;
    srcIndex.Set1DInt(m, 0);
    srcIndex.Set1DInt(n, 1);

    InitTensor2D(&a, 3, 2, X_FLOAT, 0);
    InitTensor2D(&b, 2, 2, X_FLOAT, 0);
    InitTensor2D(&halfB, 2, 2, X_FLOAT16, 0);

    a.SetDataRand(-5.0F, 5.0F);
    halfA = ConvertDataType(a, X_FLOAT16);

    a.Dump(stderr, "a:");
    _Gather(&a, &b, &srcIndex);

    b.Dump(stderr, "b:");

    _Gather(&halfA, &halfB, &srcIndex);
    c = ConvertDataType(halfB, X_FLOAT);
    c.Dump(stderr, "c:");
}

void DivFP16Test() {
    XTensor a;
    XTensor b;
    XTensor c;
    XTensor halfA;
    XTensor halfB;
    XTensor halfC;

    InitTensor2D(&a, 2, 2, X_FLOAT, 0);
    InitTensor2D(&b, 2, 2, X_FLOAT, 0);
    InitTensor2D(&c, 2, 2, X_FLOAT, 0);

    a.SetDataRand(5.0F, 5.0F);
    b.SetDataRand(2.0F, 2.0F);
    c.SetDataRand(4.0F, 4.0F);

    halfA = ConvertDataType(a, X_FLOAT16);
    halfB = ConvertDataType(b, X_FLOAT16);
    halfC = ConvertDataType(c, X_FLOAT16);

    //c = Div(a, b, 4.0F);
    //halfC = Div(halfA, halfB, 4.0F);

    _Div(&a, &b, &c, 1.0F);
    _Div(&halfA, &halfB, &halfC, 1.0F);

    c.Dump(stderr, "c:");
    halfC.Dump(&halfC, stderr, "halfC:");
}

void PowerFP16Test() {

    srand(time(NULL));
    XTensor a;
    XTensor b;
    XTensor halfa;
    XTensor halfb;

    InitTensor2D(&a, 2, 2, X_FLOAT, 0);
    InitTensor2D(&b, 2, 2, X_FLOAT, 0);
    a.SetDataRand(-4.0F, 4.0F);

    halfa = ConvertDataType(a, X_FLOAT16);

    b = Power(a, 0.5F);
    halfb = Power(halfa, 0.5F);

    b.Dump(stderr, "b:");
    halfb.Dump(&halfb, stderr, "halfb:");

}

void DivDimFP16Test() {

    XTensor a;
    XTensor b;
    XTensor c;

    XTensor c1;

    XTensor halfA;
    XTensor halfB;
    XTensor halfC;

    InitTensor2D(&a, 2, 2, X_FLOAT, 0);
    InitTensor1D(&b, 2, X_FLOAT, 0);
    InitTensor2D(&c, 2, 2, X_FLOAT, 0);

    a.SetDataRand(5.0F, 5.0F);
    c.SetDataRand(1.0F, 1.0F);
    b.Set1D(1.0F, 0);
    b.Set1D(-1.0F, 1);

    a.Dump(stderr, "a:");
    b.Dump(stderr, "b:");

    c = DivDim(a, b, 0, 0.0F);

    c1 = DivDim(a, b, 0, 0.0F);

    c.Dump(stderr, "c:");
    c1.Dump(stderr, "c1:");

    halfA = ConvertDataType(a, X_FLOAT16);
    halfB = ConvertDataType(b, X_FLOAT16);
    halfC = DivDim(halfA, halfB, 0, 0.0F);

    halfC.Dump(&halfC, stderr, "halfC:");   

}

void SumDimFP16Test() {
    printf("\nSumDim:\n");
    XTensor a;
    XTensor b;
    XTensor c;
    XTensor c1;
    XTensor halfA;
    XTensor halfB;
    XTensor halfC;
    InitTensor2D(&a, 2, 2, X_FLOAT, 0);
    InitTensor2D(&b, 1, 2, X_FLOAT, 0);

    a.SetDataRand(-5.0F, 5.0F);
    b.SetDataRand(-5.0F, 5.0F);

    a.Dump(stderr, "a:\n");
    b.Dump(stderr, "b:\n");
    printf("float:\n");

    c = SumDim(a, b, 1, 4.5);//计算float
    c.Dump(stderr, "c:\n");

    halfA = ConvertDataType(a, X_FLOAT16);
    halfB = ConvertDataType(b, X_FLOAT16);
    halfC = SumDim(halfA, halfB, 1, 4.5);//计算float16

    printf("floatHalf:\n");
    c1 = ConvertDataType(halfC, X_FLOAT);//float16 转为 float32
    c1.Dump(stderr, "c1:\n");
}

void SubDimFP16Test() {

    printf("\nSubDim:\n");
    XTensor a;
    XTensor b;
    XTensor c;
    XTensor c1;
    XTensor halfA;
    XTensor halfB;
    XTensor halfC;
    InitTensor2D(&a, 2, 2, X_FLOAT, 0);
    InitTensor2D(&b, 1, 2, X_FLOAT, 0);

    a.SetDataRand(-5.0F, 5.0F);
    b.SetDataRand(-5.0F, 5.0F);

    a.Dump(stderr, "a:");
    b.Dump(stderr, "b:");
    printf("float:\n");
    c = SubDim(a, b, 0, 3.4);//计算float
    c.Dump(stderr, "c:");

    halfA = ConvertDataType(a, X_FLOAT16);
    halfB = ConvertDataType(b, X_FLOAT16);
    halfC = SubDim(halfA, halfB, 0, 3.4);//计算float16

    printf("floatHalf:\n");
    c1 = ConvertDataType(halfC, X_FLOAT);//float16 转为 float32
    c1.Dump(stderr, "c1:");

}

void SignFP16Test() {
    XTensor a;
    XTensor b;
    XTensor halfa;
    XTensor halfb;
    InitTensor2D(&a, 3, 3, X_FLOAT, 0);
    a.SetDataRand(-5.0F, 5.0F);
    halfa = ConvertDataType(a, X_FLOAT16);
    b = Sign(a);
    halfb = Sign(halfa);
    b.Dump(stderr, "b:");
    halfb.Dump(&halfb, stderr, "halfb:");
}

void FloatToInt8Test() {
    XTensor a;
    XTensor b;
    XTensor c;
    InitTensor2D(&a, 2, 2, X_FLOAT, 0);
    InitTensor2D(&b, 2, 2, X_INT8, 0);

    a.SetDataRand(5.0F, 5.0F);
    a.Dump(stderr, "a:");

    b = ConvertDataType(a, X_INT8);
    c = ConvertDataType(b, X_FLOAT);
    c.Dump(stderr, "c:");

}

void SetDataFP16Test() {
    XTensor a;
    InitTensor2D(&a, 2, 2, X_FLOAT16, 0);
    a.SetDataRand(-1.0F, 1.0F);
    XTensor b;
    InitTensor2D(&b, 2, 2, X_FLOAT, 0);
    
}

void FloatToFloat16Test() {
    
    for (int i = 0; i <= 2; i++) {
        XTensor a;
        XTensor b;
        XTensor c;
        XTensor halfa;
        XTensor halfb;
        XTensor halfc;

        InitTensor2D(&a, 2, 2, X_FLOAT, 0);
        InitTensor2D(&b, 2, 2, X_FLOAT, 0);

        a.SetDataRand(2.0F, 2.0F);
        b.SetDataRand(2.0F, 2.0F);

        halfa = ConvertDataType(a, X_FLOAT16);
        halfb = ConvertDataType(b, X_FLOAT16);

        halfc = MatrixMul(halfb, halfb);

        halfc.Dump(&halfa, stderr, "halfc");
    }

}

void LogSoftmaxFP16Test() {
   
    XTensor a;
    XTensor a1;
    XTensor halfa;
    XTensor b;
    XTensor b1;
    XTensor halfb;

    InitTensor3D(&a, 2, 2, 2, X_FLOAT, 0);
    a.SetDataRand(-1.0F, 1.0F);
    a1 = ScaleAndShift(a, 1, 0);

    halfa = ConvertDataType(a, X_FLOAT16);

    halfb = LogSoftmax(halfa, 1);

    b = ConvertDataType(halfb, X_FLOAT);

    b1 = LogSoftmax(a1, 1);

    b.Dump(stderr, "b: ");
    b1.Dump(stderr, "b1: ");

}

void TimeTest1() {
    
    //TEST 1
    XTensor a;
    XTensor a1;
    XTensor halfa;
    XTensor b;
    XTensor b1;
    XTensor halfb;
    XTensor c;
    XTensor c1;
    XTensor halfc;

    InitTensor3D(&a, 86, 48, 256, X_FLOAT, 0);
    InitTensor2D(&b, 256, 256, X_FLOAT, 0);

    a.SetDataRand(-2.0F, 2.0F);
    b.SetDataRand(-2.0F, 2.0F);

    a1 = ScaleAndShift(a, 1, 0);
    b1 = ScaleAndShift(b, 1, 0);

    int rnum = 10000;

    double startT = GetClockSec();

    for (int i = 0; i < rnum; i++) {
        c1 = MatrixMul(a1, b1);
    }

    double elapsed = GetClockSec() - startT;

    double startT1 = GetClockSec();

    for (int i = 0; i < rnum; i++) {
        halfa = ConvertDataType(a, X_FLOAT16);
        halfb = ConvertDataType(b, X_FLOAT16);
    }  

    double elapsed1 = GetClockSec() - startT1;

    double startT2 = GetClockSec();

    for (int i = 0; i < rnum; i++) {
        halfc = MatrixMul(halfa, halfb);
    }  

    double elapsed2 = GetClockSec() - startT2;

    double startT3 = GetClockSec();

    for (int i = 0; i < rnum; i++) {
        c = ConvertDataType(halfc, X_FLOAT);
    }
       
    double elapsed3 = GetClockSec() - startT3;

    printf("%f    %f    %f    %f", elapsed, elapsed1, elapsed2, elapsed3);

}

void SmallAttentionTest() 
{
    XTensor a;

    InitTensor2D(&a, 2, 2, X_FLOAT, 0);
    a.SetDataRandn(2.0F, 2.0F);

    XTensor a1;
    XTensor halfa1;
    XTensor b;
    XTensor c;
    XTensor d;

    InitTensor2D(&b, 2, 2, X_FLOAT, 0);
    b.SetDataRandn(2.0F, 2.0F);

    //a1 = ScaleAndShift(a, 1, 0);
    //halfa1 = ConvertDataType(a1, X_FLOAT16);
    //a1 = ConvertDataType(halfa1, X_FLOAT);

    halfa1 = ConvertDataType(a, X_FLOAT16);
    a = ConvertDataType(halfa1, X_FLOAT);

    c = Sum(b, a);
    d = Sum(c, a);
    
    d.grad = NewTensor(&d);
    d.grad->SetDataRand(3.0F, 3.0F);

    XNet testFloat16;
    testFloat16.Backward(d);

    d.grad->Dump(stderr, "d.grad:");

    c.grad->Dump(stderr, "c.grad:");

    b.grad->Dump(stderr, "b.grad:");

    a.grad->Dump(stderr, "a.grad:");
}

void ScaleAndShiftFP16BackwardTest()
{
    XTensor a;
    XTensor b;
    InitTensor2D(&a, 2, 2, X_FLOAT, 0);
    b = ScaleAndShift(a, 1, 0);

    b.grad = NewTensor(&b);
    b.grad->SetDataRand(3.0F, 3.0F);

    XNet testFloat32;
    testFloat32.Backward(b);

    b.grad->Dump(stderr, "b.grad:");

    a.grad->Dump(stderr, "a.grad:");
}

void ScaleAndShiftFP16Test() {

    XTensor a;
    XTensor halfa;
    XTensor b;
    XTensor halfb;

    InitTensor2D(&a, 2, 2, X_FLOAT, 0);

    a.SetDataRand(0, 0);
    halfa = ConvertDataType(a, X_FLOAT16);

    b = ScaleAndShift(a, 1.0F, 1e-7);
    halfb = ScaleAndShift(halfa, 1.0F, 1e-7);

    b.Dump(stderr, "b:");
    halfb.Dump(&halfb, stderr, "halfb:");

}

void DropoutFP16BackwardTest()
{
    XTensor a;
    XTensor a1;
    XTensor b;
    XTensor b1;
    XTensor halfA;
    XTensor halfB;

    InitTensor2D(&a, 2, 2, X_FLOAT, 0);
    InitTensor2D(&a1, 2, 2, X_FLOAT, 0);

    a.SetDataRand(2.0F, 2.0F);
    //a1 = ScaleAndShift(a, 1, 0);
    a1.SetDataRand(2.0F, 2.0F);

    //a.Dump(stderr, "a:");
    //a1.Dump(stderr, "a1:");

    halfA = ConvertDataType(a, X_FLOAT16);

    halfB = Dropout(halfA, 0.2);
    b1 = Dropout(a1, 0.2);

    b = ConvertDataType(halfB, X_FLOAT);

    //b.Dump(stderr, "b:");
    //b1.Dump(stderr, "b1:");

    b.grad = NewTensor(&b);
    b.grad->SetDataRand(3.0F, 3.0F);

    //b.grad->Dump(stderr, "b.grad:");

    b1.grad = NewTensor(&b1);
    b1.grad->SetDataRand(3.0F, 3.0F);

    //b1.grad->Dump(stderr, "b1.grad:");

    XNet testFloat16;

    testFloat16.Backward(b);

    XNet testFloat32;

    testFloat32.Backward(b1);

    a.grad->Dump(stderr, "a.grad:");
    a1.grad->Dump(stderr, "a1.grad:");

    printf("\n");
    testFloat16.ShowNetwork(stderr, &b);
    printf("\n");
    testFloat32.ShowNetwork(stderr, &b1);

}

void MultiplyDimFP16Test()
{
    XTensor a;
    XTensor b;
    XTensor a1;
    XTensor b1;
    XTensor c;
    XTensor c1;
    XTensor halfA;
    XTensor halfB;
    XTensor halfC;
    InitTensor2D(&a, 4, 4, X_FLOAT, 0);
    InitTensor1D(&b, 4, X_FLOAT, 0);
    InitTensor2D(&a1, 4, 4, X_FLOAT, 0);
    InitTensor1D(&b1, 4, X_FLOAT, 0);

    a.SetDataRand(2.0F, 2.0F);
    b.SetDataRand(1.25F, 1.25F);
    a1.SetDataRand(2.0F, 2.0F);
    b1.SetDataRand(1.25F, 1.25F);

    a.Dump(stderr, "a:");
    b.Dump(stderr, "b:");

    halfA = ConvertDataType(a, X_FLOAT16);
    halfB = ConvertDataType(b, X_FLOAT16);

    c1 = MultiplyDim(a1, b1, 1);
    halfC = MultiplyDim(halfA, halfB, 1);

    c = ConvertDataType(halfC, X_FLOAT);

    c.Dump(stderr, "c:");
    c1.Dump(stderr, "c1:");

    c.grad = NewTensor(&c);
    c.grad->SetDataRand(3.0F, 3.0F);

    //b.grad->Dump(stderr, "b.grad:");

    c1.grad = NewTensor(&c1);
    c1.grad->SetDataRand(3.0F, 3.0F);

    //b1.grad->Dump(stderr, "b1.grad:");

    XNet testFloat16;

    testFloat16.Backward(c);

    XNet testFloat32;

    testFloat32.Backward(c1);

    a.grad->Dump(stderr, "a.grad:");
    b.grad->Dump(stderr, "b.grad:");
    a1.grad->Dump(stderr, "a1.grad:");
    b1.grad->Dump(stderr, "b1.grad:");
}

void ReduceMaxFP16Test()
{
    XTensor a;
    XTensor max;
    XTensor a1;
    XTensor max1;
    XTensor halfA;
    XTensor halfMax;
    InitTensor2D(&a, 10, 10, X_FLOAT, 0);
    InitTensor1D(&max, 10, X_FLOAT, 0);

    a.SetDataRand(-5.0F, 5.0F);
    a1 = ScaleAndShift(a, 1, 0);

    //a.Dump(stderr, "a:");
    //a1.Dump(stderr, "a1:");

    halfA = ConvertDataType(a, X_FLOAT16);

    halfMax = ReduceMax(halfA, 1);
    max1 = ReduceMax(a1, 1);

    max = ConvertDataType(halfMax, X_FLOAT);

    max.Dump(stderr, "max:");
    max1.Dump(stderr, "max1:");

}

void SoftmaxBackwardFP16Test()
{
    XTensor a;
    XTensor a1;
    XTensor b;
    XTensor b1;
    XTensor halfA;
    XTensor halfB;
    InitTensor2D(&a, 5, 5, X_FLOAT, 0);

    a.SetDataRand(-5.0F, 5.0F);
    a1 = ScaleAndShift(a, 1, 0);

    //a.Dump(stderr, "a:");
    //a1.Dump(stderr, "a1:");

    halfA = ConvertDataType(a, X_FLOAT16);

    halfB = Softmax(halfA, -1);
    b1 = Softmax(a1, -1);

    b = ConvertDataType(halfB, X_FLOAT);

    //b.Dump(stderr, "b:");
    //b1.Dump(stderr, "b1:");

    b.grad = NewTensor(&b);
    b.grad->SetDataRand(3.0F, 3.0F);

    //b.grad->Dump(stderr, "b.grad:");

    b1.grad = NewTensor(&b1);
    b1.grad->SetDataRand(3.0F, 3.0F);

    //b1.grad->Dump(stderr, "b1.grad:");

    XNet testFloat16;

    testFloat16.Backward(b);

    XNet testFloat32;

    testFloat32.Backward(b1);

    a.grad->Dump(stderr, "a.grad:");
    a1.grad->Dump(stderr, "a1.grad:");

}

void UnsqueezeFP16Test() 
{
    XTensor a;
    XTensor b;
    XTensor halfA;
    XTensor halfB;
    InitTensor2D(&a, 2, 3, X_FLOAT, 0);

    a.SetDataRand(-5.0F, 5.0F);

    a.Dump(stderr, "a:");

    halfA = ConvertDataType(a, X_FLOAT16);

    halfB = Unsqueeze(a, 0, 2);

    b = ConvertDataType(halfB, X_FLOAT);

    b.Dump(stderr, "b:");
}

void MultiplyFP16Test() 
{
    XTensor a;
    XTensor b;
    XTensor halfA;
    XTensor halfB;

    InitTensor2D(&a, 2, 2, X_FLOAT, 0);
    InitTensor2D(&b, 2, 2, X_FLOAT, 0);

    a.SetDataRand(-5.0F, 5.0F);
    b.SetDataRand(-2.0F, 2.0F);

    halfA = ConvertDataType(a, X_FLOAT16);
    halfB = ConvertDataType(b, X_FLOAT16);

    _Multiply(&a, &a, &b, 0.0F);
    _Multiply(&halfA, &halfA, &halfB, 0.0F);

    b.Dump(stderr, "b:");
    halfB.Dump(&halfB, stderr, "halfB:");
}

void SumFP16Test()
{
    XTensor a;
    XTensor b;
    XTensor halfA;
    XTensor halfB;
    InitTensor2D(&a, 5, 5, X_FLOAT, 0);
    InitTensor2D(&b, 5, 5, X_FLOAT, 0);

    a.SetDataRand(-1.0F, 4.0F);
    b.SetDataRand(-1.0F, 4.0F);

    halfA = ConvertDataType(a, X_FLOAT16);
    halfB = ConvertDataType(b, X_FLOAT16);

    b = Sum(a, b, -0.4F);
    halfB = Sum(halfA, halfB, -0.4F);

    b.Dump(stderr, "b:");
    halfB.Dump(&halfB, stderr, "halfB:");

}

void SubFP16Test()
{
    XTensor a;
    XTensor b;
    XTensor c;
    XTensor c1;
    XTensor halfA;
    XTensor halfB;
    XTensor halfC;
    InitTensor2D(&a, 2, 5, X_FLOAT, 0);
    InitTensor2D(&b, 2, 5, X_FLOAT, 0);

    a.SetDataRand(-3.0F, 3.0F);
    b.SetDataRand(-3.0F, 3.0F);

    a.Dump(stderr, "a:");
    b.Dump(stderr, "b:");

    halfA = ConvertDataType(a, X_FLOAT16);
    halfB = ConvertDataType(b, X_FLOAT16);
    halfC = Sub(halfA, halfB, -1.0F);
    c1 = Sub(a, b, -1.0F);

    c = ConvertDataType(halfC, X_FLOAT);

    c.Dump(stderr, "c:");
    c1.Dump(stderr, "c1:");

}

void DropoutFP16Test()
{
    srand(time(NULL));

    XTensor a;
    XTensor b;
    XTensor b1;
    XTensor halfA;
    XTensor halfB;
    InitTensor2D(&a, 10, 1, X_FLOAT, 0);

    a.SetDataRand(-5.0F, 5.0F);

    a.Dump(stderr, "a:");
    
    halfA = ConvertDataType(a, X_FLOAT16);

    halfB = Dropout(halfA, 0.2);
    b1 = Dropout(a, 0.2);

    halfB.Dump(&halfB, stderr, "halfB:");
    b1.Dump(&b1, stderr, "b1:");
}

void ConvertBackwardTest()
{  
    int rnum = 4;
    for (int i = 0; i <= rnum; i++)
    {
        XTensor a;
        InitTensor2D(&a, 2, 2, X_FLOAT, 0);
        a.SetDataRand(2.0F, 2.0F);
        a.Dump(stderr, "a:");

        XTensor b;
        InitTensor2D(&b, 2, 2, X_FLOAT, 0);
        b.SetDataRand(2.0F, 2.0F);
        b.Dump(stderr, "b:");

        XTensor c;

        XTensor halfA;
        XTensor halfB;
        XTensor halfC;

        halfA = ConvertDataType(a, X_FLOAT16);
        halfB = ConvertDataType(b, X_FLOAT16);

        halfC = MMul(halfA, halfB);

        c = ConvertDataType(halfC, X_FLOAT);

        c.grad = NewTensor(&c);
        c.grad->SetDataRand(3.0F, 3.0F);

        c.grad->Dump(stderr, "c.grad:");

        XNet testFloat16;

        testFloat16.Backward(c);

        a.grad->Dump(stderr, "a.grad:");
        b.grad->Dump(stderr, "b.grad:");

        printf("float16 matrixmul done\n\n");

        XTensor a1;
        InitTensor2D(&a1, 2, 2, X_FLOAT, 0);
        a1.SetDataRand(2.0F, 2.0F);
        a1.Dump(stderr, "a1:");

        XTensor b1;
        InitTensor2D(&b1, 2, 2, X_FLOAT, 0);
        b1.SetDataRand(2.0F, 2.0F);
        b1.Dump(stderr, "b1:");

        XTensor c1;

        c1 = MMul(a1, b1);

        c1.grad = NewTensor(&c1);
        c1.grad->SetDataRand(3.0F, 3.0F);
        _CopyValues(c.grad, c1.grad);

        c1.grad->Dump(stderr, "c1.grad:");

        XNet testFloat32;

        testFloat32.Backward(c1);

        a1.grad->Dump(stderr, "a1.grad:");
        b1.grad->Dump(stderr, "b1.grad:");

        printf("float32 matrixmul done\n\n");
    }
}

void ConvertTest()
{
    int rnum = 5;
    for (int i = 0; i <= rnum; i++)
    {
        XTensor a;
        InitTensor2D(&a, 2, 2, X_FLOAT, 0);

        XTensor halfa;
        InitTensor2D(&halfa, 2, 2, X_FLOAT16, 0);

        XTensor a1;
        InitTensor2D(&a1, 2, 2, X_FLOAT, 0);

        a.SetDataRand(-10.0F, 10.0F);

        a.Dump(stderr, "a:");

        halfa = ConvertDataType(a, X_FLOAT16);

        a1 = ConvertDataType(halfa, X_FLOAT);

        a1.Dump(stderr, "halfa:");
    }
}